package com.riiiver.sample.qcc5100.features.device_search

import android.Manifest
import android.annotation.SuppressLint
import android.bluetooth.le.ScanResult
import android.os.Build
import androidx.compose.foundation.clickable
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.material.AlertDialog
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.ExtendedFloatingActionButton
import androidx.compose.material.Icon
import androidx.compose.material.MaterialTheme
import androidx.compose.material.Scaffold
import androidx.compose.material.ScaffoldState
import androidx.compose.material.Text
import androidx.compose.material.TextButton
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Search
import androidx.compose.material.rememberScaffoldState
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.res.dimensionResource
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.unit.dp
import androidx.compose.ui.zIndex
import androidx.hilt.navigation.compose.hiltViewModel
import androidx.lifecycle.compose.collectAsStateWithLifecycle
import com.riiiver.sample.qcc5100.R
import com.riiiver.sample.qcc5100.router.DeviceSearchNavigationBar
import com.riiiver.sample.qcc5100.util.os.RequestPermissions
import timber.log.Timber

@Composable
fun DeviceSearchScreen(
    modifier: Modifier = Modifier,
    onDeviceClick: () -> Unit,
    viewModel: DeviceSearchViewModel = hiltViewModel(),
    scaffoldState: ScaffoldState = rememberScaffoldState()
) {
    Scaffold (
        modifier = modifier.fillMaxSize(),
        scaffoldState = scaffoldState,
        topBar = { DeviceSearchNavigationBar(
            onClickInfo = viewModel::didTapInfo,
            onClickDelete = viewModel::clearScanResults
        )},
        floatingActionButton = {
            ExtendedFloatingActionButton(
                onClick = {
                    viewModel.didTapFab()
                },
                icon = { Icon(Icons.Filled.Search, stringResource(id = R.string.scan_device) ) },
                text = { Text(text = stringResource(id = R.string.scan_device)) }
            )
        }
    ) { paddingValues ->
        val uiState by viewModel.uiState.collectAsState()
        val scanResults by viewModel.scanResults.collectAsStateWithLifecycle()

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
            RequestPermissions(permissions = listOf(Manifest.permission.POST_NOTIFICATIONS)) {
                Timber.d("## isGranted: $it ##")
            }
        }
        if (uiState.isWaiting) {
            Box(
                modifier = modifier
                    .fillMaxSize()
                    .zIndex(Float.MAX_VALUE)
                    .clickable(
                        interactionSource = remember { MutableInteractionSource() },
                        indication = null,
                        onClick = {},
                    ),
                contentAlignment = Alignment.Center
            ) {
                CircularProgressIndicator(
                    modifier = modifier.width(64.dp),
                    color = MaterialTheme.colors.secondary,
                    backgroundColor = MaterialTheme.colors.background,
                )
            }
        }
        DevicesList(
            results = scanResults,
            modifier = Modifier.padding(paddingValues),
        )
        when (uiState.event) {
            DeviceSearchEvent.ShowPermissionDialog -> {
                val permissions = mutableListOf(Manifest.permission.ACCESS_FINE_LOCATION)
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.S) {
                    permissions += Manifest.permission.BLUETOOTH_SCAN
                    permissions += Manifest.permission.BLUETOOTH_CONNECT
                }
                RequestPermissions(permissions) { isGranted ->
                    Timber.d("## isGranted: $isGranted ##")
                    viewModel.didShowPermissionDialog(isGranted.all { it.value })
                }
            }
            DeviceSearchEvent.ShowUnsupportedDialog -> {
                AlertDialog(
                    text = { Text("This device is not supported. Select the device named \"SLATE_+_L\" on the found device list.") },
                    onDismissRequest = {},
                    confirmButton = {
                        TextButton(onClick = { viewModel.didHandleEvent() }) {
                            Text("OK")
                        }
                    },
                )
            }
            DeviceSearchEvent.ShowErrorDialog -> {
                AlertDialog(
                    text = { Text("An error occurred. Please try again from the beginning.") },
                    onDismissRequest = {},
                    confirmButton = {
                        TextButton(onClick = { viewModel.didHandleEvent() }) {
                            Text("OK")
                        }
                    },
                )
            }
            DeviceSearchEvent.SegueToDetails -> {
                viewModel.didHandleEvent()
                onDeviceClick()
            }
            DeviceSearchEvent.ShowAppInfoDialog -> {
                AlertDialog(
                    text = { Text("QCC5100 Sample App\n${viewModel.getAppInfo()}") },
                    onDismissRequest = {},
                    confirmButton = {
                        TextButton(onClick = { viewModel.didHandleEvent() }) {
                            Text("OK")
                        }
                    },
                )
            }
            else -> {
                // No event to handle
            }
        }
    }
}

@Composable
private fun DevicesList(
    results: Array<ScanResult>,
    modifier: Modifier = Modifier
) {
    Column(
        modifier = modifier
            .fillMaxSize()
            .padding(horizontal = dimensionResource(id = R.dimen.horizontal_margin))
    ) {
        LazyColumn {
            items(results.toList()) { result ->
                DeviceItem(scanResult = result)
            }
        }
    }
}

@SuppressLint("MissingPermission")
@Composable
private fun DeviceItem(
    scanResult: ScanResult,
    viewModel: DeviceSearchViewModel = hiltViewModel(),
) {
    Row(
        verticalAlignment = Alignment.CenterVertically,
        modifier = Modifier
            .fillMaxWidth()
            .padding(
                horizontal = dimensionResource(id = R.dimen.horizontal_margin),
                vertical = dimensionResource(id = R.dimen.list_item_padding),
            )
            .clickable { viewModel.connectDevice(scanResult.device.address) }
    ) {
        Text(
            text = scanResult.device.name ?: "Unnamed device",
            style = MaterialTheme.typography.h6,
            modifier = Modifier.padding(start = dimensionResource(id = R.dimen.horizontal_margin))
        )
    }
}