package com.riiiver.sample.qcc5100.features.device_search

import android.bluetooth.le.ScanResult
import androidx.lifecycle.SavedStateHandle
import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import com.riiiver.sample.qcc5100.util.GetAppInfoUseCase
import com.riiiver.sample.qcc5100.util.WhileUiSubscribed
import com.riiiver.sample.qcc5100.util.ble.BleApi
import com.riiiver.sample.qcc5100.util.ble.ConnectResults
import com.riiiver.sample.qcc5100.util.ble.IBleCentralClient
import dagger.hilt.android.lifecycle.HiltViewModel
import javax.inject.Inject
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach
import kotlinx.coroutines.flow.stateIn
import timber.log.Timber

/**
 * UiState for the device search result list screen
 */
data class DeviceSearchUiState(
    val isWaiting: Boolean = false,
    val event: DeviceSearchEvent? = null,
)

enum class DeviceSearchEvent {
    ShowPermissionDialog,
    ShowUnsupportedDialog,
    ShowErrorDialog,
    SegueToDetails,
    ShowAppInfoDialog,
}

/**
 * ViewModel for the device search result list screen
 */
@HiltViewModel
class DeviceSearchViewModel @Inject constructor(
    private val getAppInfoUseCase: GetAppInfoUseCase,
    private val bleCentralClient: IBleCentralClient,
    private val savedStateHandle: SavedStateHandle
) : ViewModel() {
    private val _isWaiting = MutableStateFlow(false)
    private val _event = MutableStateFlow<DeviceSearchEvent?>(null)
    private var _scanResults = savedStateHandle.getStateFlow(SCAN_RESULTS_STATE_KEY, emptyArray<ScanResult>())
    val scanResults: StateFlow<Array<ScanResult>> = _scanResults

    init {
        bleCentralClient.data.onEach { data ->
            data.scanResults?.let { results ->
                val newArray = (_scanResults.value + results).toSet().toTypedArray()
                Timber.d("## bleCentralClient.data.onEach: scan results has been updated\n## ${newArray.size}")
                savedStateHandle[SCAN_RESULTS_STATE_KEY] = results.toTypedArray()
            }
            data.connectResult?.let { result ->
                Timber.d("## bleCentralClient.data.onEach: connectResult has been updated $result ##")
                if (result.status == ConnectResults.Idle) {
                    return@let
                }
                _isWaiting.value = false
                when (result.status) {
                    ConnectResults.Unsupported -> _event.value = DeviceSearchEvent.ShowUnsupportedDialog
                    ConnectResults.CanSegue -> _event.value = DeviceSearchEvent.SegueToDetails
                    ConnectResults.NotStartConnectGatt, ConnectResults.NotStartRequestMtu, ConnectResults.NotStartDiscoverServices, ConnectResults.NotStartSetNotifyValue -> _event.value = DeviceSearchEvent.ShowErrorDialog
                    else -> {
                        // do nothing
                    }
                }
            }
        }.launchIn(viewModelScope)
    }

    val uiState: StateFlow<DeviceSearchUiState> =
        combine(
            _isWaiting, _event
        ) { isWaiting, event ->
            DeviceSearchUiState(
                isWaiting = isWaiting,
                event = event,
            )
        }
        .stateIn(
            scope = viewModelScope,
            started = WhileUiSubscribed,
            initialValue = DeviceSearchUiState()
        )

    fun didTapFab() {
        Timber.d("## didTapFab ##")
        _event.value = DeviceSearchEvent.ShowPermissionDialog
    }

    fun didShowPermissionDialog(isGranted: Boolean) {
        Timber.d("## didShowPermissionDialog: is granted = $isGranted##")
        if (isGranted) {
            _event.value = null
            startBleScan()
        }
        // FIXME: show dialog telling that the app cannot work without the permission
        else {
            _event.value = DeviceSearchEvent.ShowPermissionDialog
        }
    }

    fun startBleScan() {
        Timber.d("## startBleScan ##")
        bleCentralClient.request(BleApi.StartScan())
    }

    fun connectDevice(deviceAddress: String) {
        Timber.d("## connectDevice ##")
        _isWaiting.value = true
        bleCentralClient.request(BleApi.Connect(deviceAddress = deviceAddress))
    }

    fun didHandleEvent() {
        _event.value = null
    }

    fun didTapInfo() {
        _event.value = DeviceSearchEvent.ShowAppInfoDialog
    }

    fun getAppInfo(): String {
        val version = getAppInfoUseCase.versionName()
        val code = getAppInfoUseCase.versionCode().toString()
        return "Version: $version ($code)"
    }

    fun clearScanResults() {
        savedStateHandle[SCAN_RESULTS_STATE_KEY] = emptyArray<ScanResult>()
        bleCentralClient.request(BleApi.ClearResults())
    }
}

const val SCAN_RESULTS_STATE_KEY = "SCAN_RESULTS_STATE_KEY"