package com.riiiver.sample.qcc5100.features.device_settings

import androidx.compose.foundation.Image
import androidx.compose.foundation.clickable
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.verticalScroll
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.DropdownMenu
import androidx.compose.material.DropdownMenuItem
import androidx.compose.material.MaterialTheme
import androidx.compose.material.Scaffold
import androidx.compose.material.ScaffoldState
import androidx.compose.material.Text
import androidx.compose.material.TextButton
import androidx.compose.material.rememberScaffoldState
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.res.dimensionResource
import androidx.compose.ui.res.painterResource
import androidx.compose.ui.unit.dp
import androidx.compose.ui.zIndex
import androidx.hilt.navigation.compose.hiltViewModel
import androidx.lifecycle.Lifecycle
import com.riiiver.sample.qcc5100.R
import com.riiiver.sample.qcc5100.router.DeviceSettingsNavigationBar
import com.riiiver.sample.qcc5100.data.AssignedFeature
import com.riiiver.sample.qcc5100.data.SlateFeature
import com.riiiver.sample.qcc5100.util.ObserveLifecycleEvent
import com.riiiver.sample.qcc5100.util.ble.CommandReceiverCommand
import timber.log.Timber

@Composable
fun DeviceSettingsScreen(
    modifier: Modifier = Modifier,
    scaffoldState: ScaffoldState = rememberScaffoldState(),
    viewModel: DeviceSettingsViewModel = hiltViewModel(),
    onBack: () -> Unit,
    onClickSettings: () -> Unit
) {
    Scaffold(
        modifier = modifier.fillMaxSize(),
        scaffoldState = scaffoldState,
        topBar = {
            DeviceSettingsNavigationBar {
                viewModel.willPop()
                onBack()
            }
        }
    ) { paddingValues ->
        val uiState by viewModel.uiState.collectAsState()

        if (uiState.isConnected) {
            DeviceSettingsContent(
                modifier = modifier.padding(paddingValues),
                loading = uiState.isLoading,
                features = uiState.features,
                assigned = uiState.assigned,
                expanded = uiState.expanded,
                onClickSettings = onClickSettings
            )
        } else {
            DisconnectedContent()
        }
    }
}

@Composable
private fun DeviceSettingsContent(
    modifier: Modifier = Modifier,
    viewModel: DeviceSettingsViewModel = hiltViewModel(),
    loading: Boolean,
    features: List<SlateFeature>,
    assigned: AssignedFeature,
    expanded: ExpandedState,
    onClickSettings: () -> Unit
) {
    if (loading) {
        Box(
            modifier = modifier
                .fillMaxSize()
                .zIndex(Float.MAX_VALUE)
                .clickable(
                    interactionSource = remember { MutableInteractionSource() },
                    indication = null,
                    onClick = {},
                ),
            contentAlignment = Alignment.Center
        ) {
            CircularProgressIndicator(
                modifier = modifier.width(64.dp),
                color = MaterialTheme.colors.secondary,
                backgroundColor = MaterialTheme.colors.background,
            )
        }
    }
    else {
        Column (
            modifier = modifier
                .fillMaxWidth()
                .padding(dimensionResource(id = R.dimen.horizontal_margin))
                .verticalScroll(rememberScrollState())
        ) {
            Image(
                painter = painterResource(id = R.drawable.button_description),
                contentDescription = null,
                modifier = modifier
                    .fillMaxWidth()
                    .size(200.dp),
                alignment = Alignment.Center,
            )
            Row(
                modifier = modifier.fillMaxWidth(),
                verticalAlignment = Alignment.CenterVertically
            ) {
                Text(text = "Function for Button1: ")
                TextButton(onClick = { viewModel.onDropdownMenuClicked(CommandReceiverCommand.BUTTON_1_RESULT) }) {
                    Text(text = assigned.button1.displayName)
                }
            }
            Row(
                modifier = modifier.fillMaxWidth(),
                verticalAlignment = Alignment.CenterVertically
            ) {
                Text(text = "Function for Button2: ")
                TextButton(onClick = { viewModel.onDropdownMenuClicked(CommandReceiverCommand.BUTTON_2_RESULT) }) {
                    Text(text = assigned.button2.displayName)
                }
            }
            DropdownMenu(
                expanded = expanded.value,
                onDismissRequest = { viewModel.onDismissDropdownMenu() },
                modifier = modifier.fillMaxWidth()
            ) {
                features.forEach { feature ->
                    DropdownMenuItem(onClick = { viewModel.onDropdownItemClicked(feature, expanded.target) }) {
                        Text(text = feature.displayName)
                    }
                }
            }
            TextButton(onClick = onClickSettings) {
                Text(text = "Add function with Riiiver")
            }
        }
    }
    ObserveLifecycleEvent { event ->
        when (event) {
            Lifecycle.Event.ON_RESUME -> {
                Timber.d("DeviceSettingsContent ON_RESUME")
                viewModel.updateFeatures()
            }
            else -> {}
        }
    }
}

@Composable
private fun DisconnectedContent(
    modifier: Modifier = Modifier
) {
    Box (
        modifier = modifier
            .fillMaxSize()
            .padding(dimensionResource(id = R.dimen.horizontal_margin))
    ) {
        Text(text = "Disconnected")
    }
}
