package com.riiiver.sample.qcc5100.features.device_settings

import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import com.riiiver.sample.qcc5100.util.WhileUiSubscribed
import com.riiiver.sample.qcc5100.data.AssignedFeature
import com.riiiver.sample.qcc5100.data.DummySlateFeature
import com.riiiver.sample.qcc5100.data.SlateFeature
import com.riiiver.sample.qcc5100.util.ble.BleApi
import com.riiiver.sample.qcc5100.util.ble.IBleCentralCallback
import com.riiiver.sample.qcc5100.util.ble.IBleCentralClient
import com.riiiver.sample.qcc5100.util.ble.CommandReceiverCommand
import com.riiiver.sample.qcc5100.util.extensions.activatedApplets
import com.riiiver.sample.qcc5100.util.extensions.kickedByQCC5100Trigger
import com.riiiver.sample.qcc5100.util.extensions.toSlateFeature
import com.riiiver.sample.qcc5100.util.riiiver.IRiiiverSdkClient
import dagger.hilt.android.lifecycle.HiltViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.launch
import timber.log.Timber
import javax.inject.Inject

/**
 * UiState for the device settings screen
 */
data class DeviceSettingsUiState(
    val isLoading: Boolean = false,
    val isConnected: Boolean = true,
    val features: List<SlateFeature> = emptyList(),
    val assigned: AssignedFeature = AssignedFeature,
    val expanded: ExpandedState = ExpandedState()
)

data class ExpandedState(
    val value: Boolean = false,
    val target: CommandReceiverCommand = CommandReceiverCommand.BUTTON_1_RESULT
)

@HiltViewModel
class DeviceSettingsViewModel @Inject constructor(
    private val bleCentralClient: IBleCentralClient,
    private val riiiverSdkClient: IRiiiverSdkClient,
) : ViewModel(), IBleCentralCallback {
    private val defaultFeatures = listOf(DummySlateFeature.ControlMusic, DummySlateFeature.RingPhone)
    private val _isConnected = MutableStateFlow(true)
    private val _isLoading = MutableStateFlow(true)
    private val _features = MutableStateFlow(defaultFeatures)
    private val _assigned = MutableStateFlow(AssignedFeature)
    private val _expanded = MutableStateFlow(ExpandedState())

    init {
        bleCentralClient.data.onEach { data ->
            data.connectResult?.gatt?.let {
                _isConnected.value = true
            } ?: run {
                Timber.d("## DISCONNECTED ##")
                _isConnected.value = false
            }
        }.launchIn(viewModelScope)

        riiiverSdkClient.data.onEach { data ->
            Timber.d("## assigned features stored in RiiiverSdkClient = ${data.assigned}, in ViewModel = ${_assigned.value}")
            _assigned.value = data.assigned
        }.launchIn(viewModelScope)
    }

    val uiState: StateFlow<DeviceSettingsUiState> = combine(
        _isConnected, _isLoading, _features, _assigned, _expanded
        ) { isConnected, isLoading, features, assigned, expanded ->
            if (isLoading) {
                DeviceSettingsUiState(isLoading = true)
            } else {
                DeviceSettingsUiState(
                    isLoading = false,
                    isConnected = isConnected,
                    features = features,
                    assigned = assigned,
                    expanded = expanded
                )
            }
        }
        .stateIn(
            scope = viewModelScope,
            started = WhileUiSubscribed,
            initialValue = DeviceSettingsUiState(isLoading = true)
        )

    fun updateFeatures() {
        if (!_isConnected.value) {
            Timber.d("## updateFeatures: device connection is lost ##")
            return
        }

        Timber.d("## updateFeatures ##")
        viewModelScope.launch {
            _isLoading.emit(true)
            val applets = riiiverSdkClient.getMyApplets()
            val features = defaultFeatures + applets.kickedByQCC5100Trigger().activatedApplets().map { it.toSlateFeature() }
            _features.emit(features)
            _isLoading.emit(false)
        }
    }

    fun willPop() {
        Timber.d("## willPop ##")
        bleCentralClient.request(BleApi.Disconnect())
    }

    fun onDropdownMenuClicked(assignFor: CommandReceiverCommand) {
        Timber.d("## onDropdownMenuClicked ##")
        when (assignFor) {
            CommandReceiverCommand.BUTTON_1_RESULT -> {
                _expanded.value = _expanded.value.copy(value = true, target = assignFor)
            }
            CommandReceiverCommand.BUTTON_2_RESULT -> {
                _expanded.value = _expanded.value.copy(value = true, target = assignFor)
            }
            else -> {
                Timber.e("## onDropdownMenuClicked: invalid argument found(assignFor = $assignFor) ##")
                throw IllegalArgumentException("Invalid argument: $assignFor")
            }
        }
    }

    fun onDismissDropdownMenu() {
        Timber.d("## onDismissDropdownMenu ##")
        _expanded.value = _expanded.value.copy(value = false)
    }

    fun onDropdownItemClicked(feature: SlateFeature, assignFor: CommandReceiverCommand) {
        Timber.d("## onDropdownItemClicked ##")
        when (assignFor) {
            CommandReceiverCommand.BUTTON_1_RESULT -> {
                riiiverSdkClient.changeTargetApplet(assignFor, feature)
            }
            CommandReceiverCommand.BUTTON_2_RESULT -> {
                riiiverSdkClient.changeTargetApplet(assignFor, feature)
            }
            else -> {
                Timber.e("## onDropdownItemClicked: invalid argument found(assignFor = $assignFor) ##")
                throw IllegalArgumentException("Invalid argument: $assignFor")
            }
        }
        onDismissDropdownMenu()
    }
}