package com.riiiver.sample.qcc5100.features.riiiver_settings

import android.webkit.WebViewClient
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.material.Scaffold
import androidx.compose.material.ScaffoldState
import androidx.compose.material.Text
import androidx.compose.material.rememberScaffoldState
import androidx.compose.runtime.Composable
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalUriHandler
import androidx.compose.ui.platform.UriHandler
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.viewinterop.AndroidView
import androidx.hilt.navigation.compose.hiltViewModel
import androidx.lifecycle.Lifecycle
import com.riiiver.sample.qcc5100.R
import com.riiiver.sample.qcc5100.router.RiiiverSettingsNavigationBar
import com.riiiver.sample.qcc5100.util.ObserveLifecycleEvent
import com.riiiver.sample.qcc5100.util.os.RequestPermission
import com.riiiver.sample.qcc5100.util.riiiver.RiiiverSdkConstants
import com.riiiver.sdk.ERPageType
import com.riiiver.sdk.ERUserWebPagesView
import timber.log.Timber

@Composable
fun RiiiverSettingsScreen(
    modifier: Modifier = Modifier,
    viewModel: RiiiverSettingsViewModel = hiltViewModel(),
    scaffoldState: ScaffoldState = rememberScaffoldState(),
    onBack: () -> Unit,
    webToken: String? = null,
) {
    Scaffold (
        modifier = modifier.fillMaxWidth(),
        scaffoldState = scaffoldState,
        topBar = { RiiiverSettingsNavigationBar (onBack) }
    ) { paddingValues ->
        val uiState by viewModel.uiState.collectAsState()

        // login is not completed
        if (!uiState.isLoggedIn) {
            NotLoggedInComponent(webToken)
        }
        // already logged in
        else {
            RiiiverSettingsComponent(
                modifier = modifier.padding(paddingValues)
            )
        }
        // receive a request via web view delegate
        val request = uiState.request
        if (request != null) {
            when (request.action) {
                // close the screen(back to the previous screen)
                UserPageDelegateAction.Close -> {
                    viewModel.didHandleRequest()
                    onBack()
                }
                // open a new url
                UserPageDelegateAction.OpenUrl -> {
                    val uriHandler = LocalUriHandler.current
                    uriHandler.openUri(request.url)
                    viewModel.didHandleRequest()
                }
                // request a permission
                UserPageDelegateAction.RequestPermission -> {
                    RequestPermission(permission = request.permission) {
                        viewModel.didHandleRequest()
                    }
                }
            }
        }
    }
}

@Composable fun NotLoggedInComponent(
    webToken: String? = null,
    uriHandler: UriHandler = LocalUriHandler.current,
    modifier: Modifier = Modifier,
    viewModel: RiiiverSettingsViewModel = hiltViewModel(),
) {
    val loginScheme: String = stringResource(id = R.string.login_scheme)

    Text(text = "Log into Riiiver...", modifier = modifier.clickable { /* openLoginScreen(viewModel) */})

    ObserveLifecycleEvent { event ->
        when (event) {
            Lifecycle.Event.ON_RESUME -> {
                Timber.d("RiiiverSettingsScreen.NotLoginComponent ON_RESUME")
                // webToken is not null (open by custom scheme)
                if (webToken != null) {
                    val redirectUri = "${loginScheme}:///weblogin?token=$webToken"
                    viewModel.verifyLogin(redirectUri)
                }
                // should open web login
                else {
                    val url = viewModel.getLoginUrl()
                    uriHandler.openUri(url.toString())
                }
            }
            else -> {}
        }
    }
}

@Composable
private fun RiiiverSettingsComponent(
    modifier: Modifier = Modifier,
    viewModel: RiiiverSettingsViewModel = hiltViewModel(),
) {
    AndroidView(
        factory = { context ->
            ERUserWebPagesView(context)
                .apply {
                    webViewClient = WebViewClient()
                    delegate = viewModel
                }.also {
                    it.loadPage(ERPageType.Top(RiiiverSdkConstants.deviceId))
                }
        },
        modifier = modifier.fillMaxSize(),
    )
}
