package com.riiiver.sample.qcc5100.features.riiiver_settings

import androidx.lifecycle.ViewModel
import androidx.lifecycle.viewModelScope
import com.riiiver.sample.qcc5100.di.ApplicationScope
import com.riiiver.sample.qcc5100.util.WhileUiSubscribed
import com.riiiver.sample.qcc5100.util.extensions.toManifestPermission
import com.riiiver.sample.qcc5100.util.riiiver.IRiiiverSdkClient
import com.riiiver.sdk.ERError
import com.riiiver.sdk.ERPermissionType
import com.riiiver.sdk.delegates.ERUserWebPagesViewDelegate
import dagger.hilt.android.lifecycle.HiltViewModel
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.stateIn
import kotlinx.coroutines.launch
import timber.log.Timber
import java.net.URL
import javax.inject.Inject

data class RiiiverSettingsUiState(
    val isLoggedIn: Boolean = false,
    val request: RequestFromUserPage? = null,
    val hasError: Boolean = false,
)

@HiltViewModel
class RiiiverSettingsViewModel @Inject constructor(
    @ApplicationScope private val externalScope: CoroutineScope,
    private val riiiverSdkClient: IRiiiverSdkClient,
) : ViewModel(), ERUserWebPagesViewDelegate {
    private val _isLoggedIn = riiiverSdkClient.isLoggedIn
    private val _request = MutableStateFlow<RequestFromUserPage?>(null)
    private val _hasError = MutableStateFlow(false)

    val uiState: StateFlow<RiiiverSettingsUiState> =
        combine(
            _isLoggedIn, _request, _hasError
        ) { isLoggedIn, request, hasError ->
            RiiiverSettingsUiState(
                isLoggedIn = isLoggedIn,
                request = request,
                hasError = hasError,
            )
        }
            .stateIn(
                scope = viewModelScope,
                started = WhileUiSubscribed,
                initialValue = RiiiverSettingsUiState()
            )

    fun getLoginUrl(): URL {
        return riiiverSdkClient.startWebLogin()
    }

    fun verifyLogin(redirectUri: String) {
        Timber.d("## verifyLogin invoked. redirectUri = $redirectUri ##")
        externalScope.launch {
            riiiverSdkClient.continueWebLogin(redirectUri)
        }
    }

    fun didHandleRequest() {
        _request.value = null
    }

    ///
    /// Overrides of ERUserWebPagesViewDelegate
    ///
    override fun didReceiveCloseRequest() {
        Timber.d("## didReceiveCloseRequest ##")
        _request.value = RequestsFromUserPage.Close()
    }

    override fun didReceiveOpenUrlRequest(url: String) {
        Timber.d("## didReceiveOpenUrlRequest: $url ##")
        _request.value = RequestsFromUserPage.OpenUrl(url = url)
    }

    override fun didReceivePermissionRequest(permissionType: ERPermissionType) {
        Timber.d("## didReceivePermissionRequest: $permissionType ##")
        val permission = permissionType.toManifestPermission()
        if (permission.isEmpty()) {
            return
        }
        _request.value = RequestsFromUserPage.RequestPermission(permission = permission)
    }

    override fun onError(error: ERError) {
        Timber.d("## onError: ${error.errorMessage} ##")
        _hasError.value = true
    }
}

enum class UserPageDelegateAction {
    OpenUrl,
    RequestPermission,
    Close,
}

interface RequestFromUserPage {
    val action: UserPageDelegateAction
    val url: String
        get() = ""
    val permission: String
        get() = ""
}

class RequestsFromUserPage {
    data class OpenUrl(
        override val action: UserPageDelegateAction = UserPageDelegateAction.OpenUrl,
        override val url: String
    ) : RequestFromUserPage

    data class RequestPermission(
        override val action: UserPageDelegateAction = UserPageDelegateAction.RequestPermission,
        override val permission: String
    ) : RequestFromUserPage

    data class Close(
        override val action: UserPageDelegateAction = UserPageDelegateAction.Close
    ) : RequestFromUserPage
}
