package com.riiiver.sample.qcc5100.util.ble

import android.annotation.SuppressLint
import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothGatt
import android.bluetooth.BluetoothGattCharacteristic
import android.bluetooth.BluetoothGattDescriptor
import android.bluetooth.BluetoothManager
import android.bluetooth.le.ScanFilter
import android.bluetooth.le.ScanResult
import android.bluetooth.le.ScanSettings
import android.content.Context
import android.os.Build
import android.os.Handler
import android.os.Looper
import com.riiiver.sample.qcc5100.data.AssignedFeature
import com.riiiver.sample.qcc5100.data.QCC5100Queue
import com.riiiver.sample.qcc5100.di.ApplicationScope
import com.riiiver.sample.qcc5100.util.extensions.onBoardUserEventCommand
import com.riiiver.sample.qcc5100.util.extensions.data
import com.riiiver.sample.qcc5100.util.extensions.subData
import com.riiiver.sample.qcc5100.util.extensions.toHexString
import com.riiiver.sample.qcc5100.util.extensions.toSlateRiiiverPayload
import com.riiiver.sample.qcc5100.util.riiiver.piece_core.QCC5100TriggerPieceCore
import dagger.hilt.android.qualifiers.ApplicationContext
import kotlinx.coroutines.CoroutineScope
import java.util.ArrayList
import java.util.UUID
import javax.inject.Inject
import javax.inject.Singleton
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.MutableSharedFlow
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach
import kotlinx.coroutines.flow.shareIn
import timber.log.Timber

enum class ConnectResults {
    Idle,
    NotStartConnectGatt,
    NotStartRequestMtu,
    NotStartDiscoverServices,
    NotStartSetNotifyValue,
    Unsupported,
    CanSegue,
}

data class ConnectResult(
    val gatt: BluetoothGatt? = null,
    val status: ConnectResults = ConnectResults.Idle,
)

data class BleCentralClientData(
    val scanResults: Set<ScanResult>? = null,
    val connectResult: ConnectResult? = null,
)

interface IBleCentralClient {
    fun request(operation: BleOperation)
    val data: Flow<BleCentralClientData>
}

@Singleton
@SuppressLint("MissingPermission")
class BleCentralClient @Inject constructor(
    @ApplicationContext private val appContext: Context,
    @ApplicationScope private val externalScope: CoroutineScope,
    private val scanCallback: IBleScanCallback,
    private val gattCallback: IBleGattCallback,
) : IBleCentralClient, IBleCentralCallback {
    private val btManager: BluetoothManager = appContext.getSystemService(BluetoothManager::class.java)
    private val bleScanner = btManager.adapter.bluetoothLeScanner
    private val backgroundHandler: Handler = Handler(Looper.getMainLooper())
    private var isScanning = false

    private val _scanResults = MutableSharedFlow<Set<ScanResult>?>(replay = 1)
    private val _connectResult = MutableStateFlow(ConnectResult())

    override val data: SharedFlow<BleCentralClientData> = combine(
        _scanResults, _connectResult
    ) { scanResults, connectResult ->
        BleCentralClientData(
            scanResults = scanResults,
            connectResult = connectResult,
        )
    }.shareIn(
        scope = externalScope,
        started = SharingStarted.WhileSubscribed(),
        replay = 1,
    )

    companion object {
        var sharedInstance: IBleCentralClient? = null
            private set
    }

    init {
        sharedInstance = this

        gattCallback.setCallback(this)

        scanCallback.resultsFlow.onEach { results ->
            _scanResults.tryEmit(results)
        }.launchIn(externalScope)
    }

    override fun request(operation: BleOperation) {
        when (operation.type) {
            Operation.START_SCAN -> startBleScan()
            Operation.STOP_SCAN -> stopBleScan()
            Operation.CONNECT -> {
                operation.deviceAddress?.let { address ->
                    connectDevice(address)
                } ?: run {
                    throw IllegalArgumentException("Device address is null")
                }
            }
            Operation.DISCONNECT -> disconnectDevice()
            Operation.DISCOVER_SERVICES -> discoverServices()
            Operation.SET_NOTIFICATION -> setNotifyValue()
            Operation.WRITE_CHARACTERISTIC -> {
                if (operation.message == null || operation.command == null) {
                    throw IllegalArgumentException("Message amd/or command is null")
                } else {
                    writeValue(genPayload(operation.message!!, operation.command!!))
                }
            }
            Operation.CLEAR_RESULTS -> clearScanResults()
        }
    }

    private fun startBleScan(timeoutPeriod: Long = 5000) {
        if (isScanning) {
            Timber.d("Already scanning")
            return
        }

        // Stops scanning after a pre-defined scan period.
        backgroundHandler.postDelayed({
            stopBleScan()
        }, timeoutPeriod)

        // Start scanning
        Timber.i("## BLE SCAN START ##")
        isScanning = true
        bleScanner.startScan(
            buildScanFilters(),
            buildScanSettings(),
            scanCallback
        )
    }

    /**
     * Return a List of [android.bluetooth.le.ScanFilter] objects to filter by Service UUID.
     */
    private fun buildScanFilters(): List<ScanFilter> {
        val scanFilters: MutableList<ScanFilter> = ArrayList()
        val builder = ScanFilter.Builder()
        builder.setServiceUuid(null)
        scanFilters.add(builder.build())
        return scanFilters
    }

    /**
     * Return a [android.bluetooth.le.ScanSettings] object set to use low power (to preserve
     * battery life).
     */
    private fun buildScanSettings(): ScanSettings {
        val builder = ScanSettings.Builder()
        builder.setScanMode(ScanSettings.SCAN_MODE_LOW_POWER)
        return builder.build()
    }

    private fun stopBleScan() {
        if (!isScanning) {
            return
        }
        bleScanner.stopScan(scanCallback)
        isScanning = false
        Timber.i("## BLE SCAN END ##")
    }

    private fun clearScanResults() {
        scanCallback.clear()
    }

    private fun connectDevice(deviceAddress: String) {
        stopBleScan()
        try {
            val device = btManager.adapter.getRemoteDevice(deviceAddress)
            val connected = device.connectGatt(appContext,false, gattCallback, BluetoothDevice.TRANSPORT_LE)
            _connectResult.value = ConnectResult(gatt = connected)
        } catch (e: IllegalArgumentException) {
            Timber.e("Could not get remote device due to invalid address: $e")
            _connectResult.value = _connectResult.value.copy(status = ConnectResults.NotStartConnectGatt)
        }
    }

    private fun disconnectDevice() {
        _connectResult.value.gatt?.disconnect()
    }

    private fun closeGatt() {
        _connectResult.value.gatt?.close()
        _connectResult.value = _connectResult.value.copy(gatt = null, status = ConnectResults.Idle)
    }

    private fun expandMtu() {
        _connectResult.value.gatt?.requestMtu(512) ?: run {
            Timber.e("Could not start requestMtu as gatt is null ##")
            _connectResult.value = _connectResult.value.copy(status = ConnectResults.NotStartRequestMtu)
        }
    }

    /**
     * Discovers services offered by a remote device as well as their
     * characteristics and descriptors.
     */
    private fun discoverServices() {
        _connectResult.value.gatt?.discoverServices() ?: run {
            Timber.e("Could not start discoverServices as gatt is null ##")
            _connectResult.value = _connectResult.value.copy(status = ConnectResults.NotStartDiscoverServices)
        }
    }

    @Suppress("DEPRECATION")
    private fun setNotifyValue() {
        Timber.d("## setNotifyValue ##")
        _connectResult.value.gatt?.let {
            val service = it.getService(UUID.fromString(PERIPHERAL_RIIIVER_SERVICE_UUID))
            val characteristic = service.getCharacteristic(UUID.fromString(ON_BOARD_USER_EVENT_CHARACTERISTIC_UUID))
            val descriptor = characteristic.getDescriptor(UUID.fromString(CCD_UUID))
            try {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
                    it.writeDescriptor(
                        descriptor,
                        BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE
                    )
                }
                // FIXME: remove this else block when it's no longer needed i.e. stop supporting Android 12 and below
                else {
                    descriptor.value = BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE
                    it.writeDescriptor(descriptor)
                }
                it.setCharacteristicNotification(characteristic, true)
            }
            // FIXME: rethrow the exception to the caller?
            catch (e: IllegalArgumentException) {
                Timber.e("Could not write value to the descriptor: $e")
            }
        } ?: run {
            Timber.e("## setNotifyValue: gatt is null ##")
            _connectResult.value = _connectResult.value.copy(status = ConnectResults.NotStartSetNotifyValue)
        }
    }

    private fun genPayload(message: String, command: CommandReceiverCommand): ByteArray {
        // dataLength: 2 bytes, commandId: 1 byte, message: n bytes, where dataLength is not including itself.
        val dataLength = (message.toByteArray().size + 1).data().subData(0,2).reversed().toByteArray()
        val commandId = command.rawValue.toByte()
        val payload = SlateRiiiverPayload(dataLength, commandId, message.toByteArray())
        return payload.combine()
    }

    @Suppress("DEPRECATION")
    private fun writeValue(value: ByteArray) {
        Timber.d("## writeValue: ${value.toHexString()} ##")
        _connectResult.value.gatt?.let {
            val service = it.getService(UUID.fromString(PERIPHERAL_RIIIVER_SERVICE_UUID))
            val characteristic = service.getCharacteristic(UUID.fromString(COMMAND_RECEIVER_CHARACTERISTIC_UUID))
            try {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
                    it.writeCharacteristic(
                        characteristic,
                        value,
                        BluetoothGattCharacteristic.WRITE_TYPE_DEFAULT
                    )
                }
                // FIXME: remove this else block when it's no longer needed i.e. stop supporting Android 12 and below
                else {
                    characteristic.value = value
                    it.writeCharacteristic(characteristic)
                }
            }
            // FIXME: rethrow the exception to the caller?
            catch (e: IllegalArgumentException) {
                Timber.e("## writeValue: Could not write value to the characteristic: $e")
            }
        } ?: run {
            Timber.e("## writeValue: gatt is null")
        }
    }

    ///
    /// Overrides of IBleCentralCallback
    ///
    override fun onConnectionStateChange(gatt: BluetoothGatt, status: Int, newState: Int) {
        if (status == BluetoothGatt.GATT_SUCCESS) {
            if (newState == BluetoothGatt.STATE_CONNECTED) {
                Timber.i("## GATT_CONNECTED ##")
                expandMtu()
            } else if (newState == BluetoothGatt.STATE_DISCONNECTED) {
                Timber.i("## GATT_DISCONNECTED ##")
                closeGatt()
                return
            }
        } else if (status == 133) {
            // As the connection sometimes fails with status 133, we should retry the connection.
            Timber.i("## 133 STATUS ERROR. RETRY TO CONNECT. ##")
            connectDevice(gatt.device.address)
        } else {
            Timber.i("## Unhandled connection state: $status => $newState. Consider implementing something. ##")
        }
    }

    override fun onMtuChanged(gatt: BluetoothGatt, mtu: Int, status: Int) {
        // Start discovering services after the MTU request is completed.
        Timber.i("## onMtuChanged: mtu = $mtu, status = $status ##")
        discoverServices()
    }

    override fun onServicesDiscovered(gatt: BluetoothGatt, status: Int) {
        if (gatt.services.any { service -> service.uuid.toString() == PERIPHERAL_RIIIVER_SERVICE_UUID }) {
            Timber.i("## slate_riiiver_service found ##")
            _connectResult.value = _connectResult.value.copy(status = ConnectResults.CanSegue)
            setNotifyValue()
        } else {
            Timber.i("## slate_riiiver_service not found ##")
            _connectResult.value = _connectResult.value.copy(status = ConnectResults.Unsupported)
            disconnectDevice()
        }
    }

    override fun onCharacteristicChanged(
        gatt: BluetoothGatt,
        characteristic: BluetoothGattCharacteristic,
        value: ByteArray
    ) {
        val receivedPayload = value.toSlateRiiiverPayload()
        when (val w2pCommand = receivedPayload.commandId.toInt().onBoardUserEventCommand()) {
            OnBoardUserEventCommand.BUTTON_1_PRESSED, OnBoardUserEventCommand.BUTTON_2_PRESSED -> {
                Timber.i("onCharacteristicChanged() [INF] commandId:${w2pCommand.rawValue} = ${w2pCommand.name}")
                try {
                    QCC5100Queue.enqueue(w2pCommand.responseCommand)
                } catch (e: IllegalStateException) {
                    Timber.e("## enqueue failed due to ${e.localizedMessage} ##")
                    return
                }
                try {
                    val appletId = AssignedFeature.getFeatureFor(w2pCommand.responseCommand).id
                    QCC5100TriggerPieceCore.didTrigger(appletId, null)
                } catch (e: IllegalStateException) {
                    Timber.e("## didTrigger failed due to ${e.localizedMessage} ##")
                    writeValue(genPayload("Error", CommandReceiverCommand.FAILURE))
                    QCC5100Queue.dequeue()
                }
            }
            OnBoardUserEventCommand.SUCCESS, OnBoardUserEventCommand.FAILURE -> {
                Timber.i("onCharacteristicChanged() [INF] commandId:${w2pCommand.rawValue} = ${w2pCommand.name}")
            }
            else -> {
                Timber.e("onCharacteristicChanged() [INF] commandId:unknown")
            }
        }
    }
}
