package com.riiiver.sample.qcc5100.util.ble

import android.bluetooth.BluetoothGatt
import android.bluetooth.BluetoothGattCallback
import android.bluetooth.BluetoothGattCharacteristic
import com.riiiver.sample.qcc5100.util.extensions.toHexString
import timber.log.Timber
import javax.inject.Inject

abstract class IBleGattCallback: BluetoothGattCallback() {
    abstract fun setCallback(callback: IBleCentralCallback)
    abstract fun removeCallback(callback: IBleCentralCallback)
}

/**
 * Callback for GATT Writing
 */
class BleGattCallback @Inject constructor(): IBleGattCallback() {
    private var callbacks: MutableSet<IBleCentralCallback> = mutableSetOf()

    companion object {
        var sharedInstance: IBleGattCallback? = null
            private set
    }

    init {
        sharedInstance = this
    }

    /**
     * Sets an instance of [IBleCentralCallback] to receive events.
     *
     * @param callback The new callback or null, if the callback should be unregistered.
     */
    override fun setCallback(callback: IBleCentralCallback) {
        this.callbacks.add(callback)
    }

    ///
    ///  Overrides of BluetoothGattCallback
    ///
    override fun removeCallback(callback: IBleCentralCallback) {
        this.callbacks.remove(callback)
    }

    override fun onConnectionStateChange(gatt: BluetoothGatt, status: Int, newState: Int) {
        Timber.i("onConnectionStateChange() [INF] status:${status} newState:${newState}")
        if (callbacks.isNotEmpty()) {
            callbacks.forEach { it.onConnectionStateChange(gatt, status, newState) }
        }
    }

    override fun onServicesDiscovered(gatt: BluetoothGatt, status: Int) {
        Timber.i("onServicesDiscovered() [INF] status:${status}")
        if (callbacks.isNotEmpty()) {
            callbacks.forEach { it.onServicesDiscovered(gatt, status) }
        }
    }

    override fun onMtuChanged(gatt: BluetoothGatt, mtu: Int, status: Int) {
        Timber.i("onMtuChanged() [INF] mtu=$mtu status=${status}")
        if (callbacks.isNotEmpty()) {
            callbacks.forEach { it.onMtuChanged(gatt, mtu, status) }
        }
    }

    override fun onCharacteristicRead(gatt: BluetoothGatt,
                                      characteristic: BluetoothGattCharacteristic,
                                      status: Int) {
        Timber.i("onServicesDiscovered() [INF] status:${status}")
        if (callbacks.isNotEmpty()) {
            callbacks.forEach { it.onCharacteristicRead(gatt, characteristic, status) }
        }
    }

    override fun onCharacteristicWrite(gatt: BluetoothGatt,
                                       characteristic: BluetoothGattCharacteristic,
                                       status: Int) {
        Timber.i("onCharacteristicWrite() [INF] status:${status}")
        if (callbacks.isNotEmpty()) {
            callbacks.forEach { it.onCharacteristicWrite(gatt, characteristic, status) }
        }
    }

    // Though this method is deprecated, it is still called when debugging on Moto G8 running w/ Android 11.
    // FIXME: Remove this method when you stop supporting Android 12 or below.
    @Suppress("DEPRECATION")
    override fun onCharacteristicChanged(gatt: BluetoothGatt,
                                         characteristic: BluetoothGattCharacteristic,
    ) {
        Timber.d("[Deprecated API] onCharacteristicChanged() [INF] characteristic:${characteristic.uuid}")

        characteristic.value?.let { value ->
            Timber.d("[Deprecated API] onCharacteristicChanged() [INF] value:${value.toHexString()}")
            if (callbacks.isNotEmpty()) {
                callbacks.forEach { it.onCharacteristicChanged(gatt, characteristic, value) }
            }
        }
    }

    // This method is not called when debugging on Moto G8 running w/ Android 11.
    // May be called when running on Android 13 (TIRAMISU) or later as the API `writeDescriptor`
    // requires Android 13 (TIRAMISU) or later. It's just a guess tho.
    override fun onCharacteristicChanged(gatt: BluetoothGatt,
                                         characteristic: BluetoothGattCharacteristic,
                                         value: ByteArray
    ) {
        Timber.d("onCharacteristicChanged() [INF] characteristic:${characteristic.uuid}, value:${value.toHexString()}")
        if (callbacks.isNotEmpty()) {
            callbacks.forEach { it.onCharacteristicChanged(gatt, characteristic, value) }
        }
    }
}