package com.riiiver.sample.qcc5100.util.ble

import android.bluetooth.le.ScanCallback
import android.bluetooth.le.ScanResult
import android.bluetooth.le.ScanSettings
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.MutableSharedFlow
import timber.log.Timber
import javax.inject.Inject

abstract class IBleScanCallback: ScanCallback() {
    abstract val resultsFlow: Flow<Set<ScanResult>>
    abstract fun clear()
}


/**
 * Helper class for BLE scan callback.
 */
class BleScanCallback @Inject constructor() : IBleScanCallback() {
    private val _results = mutableSetOf<ScanResult>()
    private val _resultsFlow = MutableSharedFlow<Set<ScanResult>>(replay = 1)
    override val resultsFlow: Flow<Set<ScanResult>> = _resultsFlow

    /**
     * Clear scan results.
     */
    override fun clear() {
        _results.clear()
    }

    ///
    ///  Overrides of ScanCallback
    ///
    /**
     * Invoked when a Bluetooth LE peripheral has been found.
     * In this app, we only care about devices with a name.
     * */
    override fun onScanResult(callbackType: Int, result: ScanResult) {
        if (callbackType == ScanSettings.CALLBACK_TYPE_ALL_MATCHES) {
            // Check if the device has a name
            result.scanRecord?.deviceName?.let { name ->
                // Check if the device is already in the set
                if (_results.any { it.device.address == result.device.address }) { return }
                Timber.i("onScanResult() [INF] found new device which has name:$name, full length result:$result")
                _results.add(result)
                _resultsFlow.tryEmit(_results).let { isSuccess ->
                    Timber.d("onScanResult() [INF] try emmit is success: $isSuccess")
                }
            }
        }
    }
}