package com.riiiver.sample.qcc5100.util.os

import android.app.Activity
import android.content.pm.PackageManager
import androidx.activity.compose.rememberLauncherForActivityResult
import androidx.activity.result.contract.ActivityResultContracts
import androidx.compose.material.AlertDialog
import androidx.compose.material.Text
import androidx.compose.material.TextButton
import androidx.compose.runtime.Composable
import androidx.compose.runtime.DisposableEffect
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.platform.LocalLifecycleOwner
import androidx.core.app.ActivityCompat.shouldShowRequestPermissionRationale
import androidx.lifecycle.Lifecycle
import androidx.lifecycle.LifecycleEventObserver
import timber.log.Timber

@Composable
fun RequestPermissions(
    permissions: List<String>,
    onPermissionResult: (Map<String, Boolean>) -> Unit
) {
    val context = LocalContext.current
    val lifecycleOwner = LocalLifecycleOwner.current
    val launcher = rememberLauncherForActivityResult(ActivityResultContracts.RequestMultiplePermissions()) { isGranted ->
        Timber.d("permission is granted?: $isGranted")
        onPermissionResult(isGranted)
    }

    when {
        permissions.all { context.checkSelfPermission(it) == PackageManager.PERMISSION_GRANTED } -> {
            onPermissionResult(permissions.associateWith { true })
        }
        else -> {
            val shouldShowRationale = permissions.filter { shouldShowRequestPermissionRationale(context as Activity, it) }
            if (shouldShowRationale.isNotEmpty()) {
                PermissionRationaleDialog {
                    launcher.launch(shouldShowRationale.toTypedArray()) }
            }
            else {
                DisposableEffect(lifecycleOwner) {
                    val lifecycleObserver = LifecycleEventObserver { _, event ->
                        when (event) {
                            Lifecycle.Event.ON_START -> launcher.launch(permissions.toTypedArray())
                            else -> {}
                        }
                    }
                    lifecycleOwner.lifecycle.addObserver(lifecycleObserver)
                    onDispose {
                        lifecycleOwner.lifecycle.removeObserver(lifecycleObserver)
                    }
                }
            }
        }
    }

}

@Composable
fun RequestPermission(
    permission: String,
    onPermissionResult: (Boolean) -> Unit
) {
    val context = LocalContext.current
    val lifecycleOwner = LocalLifecycleOwner.current
    val launcher = rememberLauncherForActivityResult(ActivityResultContracts.RequestPermission()) { isGranted ->
            Timber.d("permission: $permission, isGranted: $isGranted")
            onPermissionResult(isGranted)
        }

    when (context.checkSelfPermission(permission)) {
        PackageManager.PERMISSION_GRANTED -> {
            onPermissionResult(true)
        }
        PackageManager.PERMISSION_DENIED -> {
            if (shouldShowRequestPermissionRationale(context as Activity, permission)) {
                PermissionRationaleDialog {
                    launcher.launch(permission)
                }
            } else {
                DisposableEffect(lifecycleOwner) {
                    val lifecycleObserver = LifecycleEventObserver { _, event ->
                        when (event) {
                            Lifecycle.Event.ON_START -> launcher.launch(permission)
                            else -> {}
                        }
                    }
                    lifecycleOwner.lifecycle.addObserver(lifecycleObserver)
                    onDispose {
                        lifecycleOwner.lifecycle.removeObserver(lifecycleObserver)
                    }
                }
            }
        }
        else -> {}
    }
}

@Composable
private fun PermissionRationaleDialog(onDialogResult: () -> Unit) {
    AlertDialog(
        text = { Text("This app would use Bluetooth to scan and connect to slate_riiiver(QCC5100 evaluation board)") },
        onDismissRequest = {},
        confirmButton = {
            TextButton(onClick = onDialogResult) {
                Text("OK")
            }
        },
    )
}