package com.riiiver.sample.qcc5100.util.riiiver

import android.content.Context
import android.net.Uri
import com.google.firebase.messaging.RemoteMessage
import com.riiiver.sample.qcc5100.R
import com.riiiver.sample.qcc5100.di.ApplicationScope
import com.riiiver.sample.qcc5100.data.AssignedFeature
import com.riiiver.sample.qcc5100.data.QCC5100Queue
import com.riiiver.sample.qcc5100.data.SlateFeature
import com.riiiver.sample.qcc5100.util.ble.BleApi
import com.riiiver.sample.qcc5100.util.ble.BleCentralClient
import com.riiiver.sample.qcc5100.util.ble.CommandReceiverCommand
import com.riiiver.sdk.ERApplet
import com.riiiver.sdk.ERBlockName
import com.riiiver.sdk.ERError
import com.riiiver.sdk.ERResult
import com.riiiver.sdk.RiiiverSdkManager
import com.riiiver.sdk.delegates.ERSdkListener
import dagger.hilt.android.qualifiers.ApplicationContext
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.combine
import kotlinx.coroutines.flow.shareIn
import kotlinx.coroutines.launch
import timber.log.Timber
import java.net.URL
import javax.inject.Inject
import javax.inject.Singleton
import kotlin.jvm.Throws

@Singleton
class RiiiverSdkClient @Inject constructor(
    @ApplicationContext private val context: Context,
    @ApplicationScope private val externalScope: CoroutineScope,
) : IRiiiverSdkClient(), ERSdkListener {
    private val _applets = MutableStateFlow(emptyList<ERApplet>())
    private val _assigned = MutableStateFlow(AssignedFeature)

    override val data: SharedFlow<RiiiverSdkClientData> = combine(
        _applets, _assigned
    ) { applets, assigned ->
        RiiiverSdkClientData(
            applets = applets,
            assigned = assigned
        )
    }.shareIn(
        scope = externalScope,
        started = SharingStarted.WhileSubscribed(),
        replay = 1,
    )

    init {
        initialize()
    }

    private fun initialize() {
        Timber.i("initialize")
        val listener: ERSdkListener = this
        externalScope.launch {
            RiiiverSdkManager.addListener(listener)
            val result = RiiiverSdkManager.initialize(
                context = context.applicationContext,
                blockExecutorPackageName = RiiiverSdkConstants.piececorePackageName,
                configResourceId = R.raw.erconfiguration,
                deviceIdList = RiiiverSdkConstants.deviceIds,
            )
            when (result) {
                is ERResult.Success -> {
                    Timber.i("initialize: success")
                }
                is ERResult.Error -> {
                    Timber.e("Error: ${result.error.errorMessage}")
                }
            }
        }
    }

    private suspend fun forcedDownloadApplets(): ERError? {
        Timber.d("forcedDownloadApplets: ${RiiiverSdkConstants.forcedDownloadAppletIds}")
        return when (val result = RiiiverSdkManager.forcedDownloadApplet(RiiiverSdkConstants.forcedDownloadAppletIds)) {
            is ERResult.Success -> {
                Timber.i("forcedDownloadApplets: success")
                null
            }

            is ERResult.Error -> {
                Timber.e("Error: ${result.error.errorMessage}")
                result.error
            }
        }
    }

    ///
    ///  Implementation of IErAuthClient
    ///
    private val _isLoggedIn = MutableStateFlow(false)
    override val isLoggedIn: Flow<Boolean>
        get() {
            return _isLoggedIn
        }

    override fun startWebLogin(): URL {
        when (val urlOrError = RiiiverSdkManager.getLoginUrl()) {
            is ERResult.Success -> {
                Timber.i("startWebLogin: success")
                return urlOrError.data
            }
            is ERResult.Error -> {
                Timber.e("Error: ${urlOrError.error.errorMessage}")
                throw urlOrError.error
            }
        }
    }

    override suspend fun continueWebLogin(uriString: String) {
        val uri = Uri.parse(uriString)
        when (val result = RiiiverSdkManager.verifyCodeOf(uri)) {
            is ERResult.Error -> {
                Timber.e("Error: ${result.error.errorMessage}")
                throw result.error
            }
            is ERResult.Success -> {
                // Nothing to do
            }
        }
    }

    override suspend fun setNotificationToken(fcmToken: String) {
        when (val result = RiiiverSdkManager.setNotificationToken(fcmToken)) {
            is ERResult.Success -> {
                Timber.i("setNotificationToken: success")
            }
            is ERResult.Error -> {
                // Not throw error
                Timber.e("Error: ${result.error.errorMessage}")
            }
        }
    }

    override suspend fun logout() {
        when (val result = RiiiverSdkManager.logout()) {
            is ERResult.Error -> {
                Timber.e("Error: ${result.error.errorMessage}")
                throw result.error
            }
            is ERResult.Success -> {
                // Nothing to do
            }
        }
    }

    ///
    ///  Implementation of IRiiiverSdkClient
    ///
    override suspend fun getMyApplets(): List<ERApplet> {
        if (!RiiiverSdkManager.authenticated) {
            return emptyList()
        }
        forcedDownloadApplets()?.let { error ->
            // Even if the forced download fails, continue to get the applets
            Timber.e("Error: $error")
        }

        return when (val result = RiiiverSdkManager.getMyApplets(RiiiverSdkConstants.deviceIds)) {
            is ERResult.Success -> {
                Timber.i("getMyApplets: success")
                _applets.value = result.data
                result.data
            }

            is ERResult.Error -> {
                // If getMyApplets fails, not throw error but return the current applets
                Timber.e("Error: ${result.error.errorMessage}")
                _applets.value
            }
        }
    }

    @Throws(IllegalArgumentException::class)
    override fun getErAppletInstance(id: String): ERApplet {
        _applets.value.firstOrNull { it.metaData.id == id }.let { applet ->
            return applet ?: throw IllegalArgumentException("Invalid applet Id")
        }
    }

    @Throws(IllegalArgumentException::class)
    override fun changeTargetApplet(of: CommandReceiverCommand, feature: SlateFeature) {
        when (of) {
            CommandReceiverCommand.BUTTON_1_RESULT -> {
                AssignedFeature.setButton1(feature)
            }
            CommandReceiverCommand.BUTTON_2_RESULT -> {
                AssignedFeature.setButton2(feature)
            }
            else -> {
                Timber.e("Invalid argument: $of")
                throw IllegalArgumentException("Invalid argument: $of")
            }
        }
    }

    override suspend fun executeTaskForRemoteNotification(remoteNotification: RemoteMessage) {
        when (val result = RiiiverSdkManager.executeTaskForRemoteNotification(remoteNotification.data)) {
            is ERResult.Success -> {
                Timber.i("executeTaskForRemoteNotification: success")
            }
            is ERResult.Error -> {
                // Not throw error
                Timber.e("Error: ${result.error.errorMessage}")
            }
        }
    }

    ///
    /// Implementation of ERSdkListener
    ///
    override fun startApplet(applet: ERApplet, error: ERError?) {
    }

    override fun endApplet(
        applet: ERApplet,
        blockName: ERBlockName,
        output: Map<String, Any>?,
        error: ERError?
    ) {
        error?.let {
            Timber.e("Error: ${it.errorMessage}")
        }
        when (val queue = QCC5100Queue.current) {
            CommandReceiverCommand.BUTTON_1_RESULT -> {
                if (error != null) {
                    val outputString = "iiidea failed due to ${error.errorMessage}"
                    BleCentralClient.sharedInstance?.request(BleApi.SendResultButton1(message = outputString))
                }
                QCC5100Queue.dequeue()
            }
            CommandReceiverCommand.BUTTON_2_RESULT -> {
                if (error != null) {
                    val outputString = "iiidea failed due to ${error.errorMessage}"
                    BleCentralClient.sharedInstance?.request(BleApi.SendResultButton2(message = outputString))
                }
                QCC5100Queue.dequeue()
            }
            CommandReceiverCommand.DISPLAY -> {
                QCC5100Queue.dequeue()
            }
            else -> {
                Timber.i("Nothing to do for the queue: $queue")
            }
        }
    }

    override fun onLogin(applets: List<ERApplet>) {
        Timber.i("onLogin: $applets")
        externalScope.launch {
            _isLoggedIn.emit(true)
            forcedDownloadApplets()
        }
    }

    override fun onLogout(error: ERError?) {
    }
}