package com.riiiver.sample.qcc5100.util.riiiver.piece_core

import android.bluetooth.BluetoothGatt
import android.bluetooth.BluetoothGattCharacteristic
import android.os.Handler
import android.os.Looper
import com.riiiver.sample.qcc5100.data.QCC5100Queue
import com.riiiver.sample.qcc5100.util.ble.BleApi
import com.riiiver.sample.qcc5100.util.ble.BleCentralClient
import com.riiiver.sample.qcc5100.util.ble.BleGattCallback
import com.riiiver.sample.qcc5100.util.ble.BleOperation
import com.riiiver.sample.qcc5100.util.ble.IBleCentralCallback
import com.riiiver.sample.qcc5100.util.ble.CommandReceiverCommand
import com.riiiver.sample.qcc5100.util.ble.OnBoardUserEventCommand
import com.riiiver.sample.qcc5100.util.extensions.onBoardUserEventCommand
import com.riiiver.sample.qcc5100.util.extensions.toSlateRiiiverPayload
import com.riiiver.sdk.ERApplet
import com.riiiver.sdk.ERBlockName
import com.riiiver.sdk.ERError
import com.riiiver.sdk.delegates.ERBlockDelegate
import com.riiiver.sdk.delegates.ERBlockExecutor
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.launch
import timber.log.Timber

class QCC5100ActionPieceCore: ERBlockExecutor, IBleCentralCallback {
    private var externalScope: CoroutineScope = CoroutineScope(SupervisorJob() + Dispatchers.IO)
    private var identifier: String? = null
    private var output: Map<String, Any> = emptyMap()
    private var handler: ERBlockDelegate? = null
    private var isRunning: Boolean = false
    private val backgroundHandler: Handler = Handler(Looper.getMainLooper())
    private val timeoutPeriod: Long = 10000

    ///
    /// Overrides of ERBlockExecutor
    ///
    override fun disable(applet: ERApplet) {
        return
    }

    override fun enable(applet: ERApplet): Boolean {
        return true
    }

    override suspend fun execute(
        applet: ERApplet,
        blockName: ERBlockName,
        preferenceJson: Map<String, Any>?,
        inputJson: Map<String, Any>?,
        handler: ERBlockDelegate
    ) {
        this.identifier = applet.metaData.id
        this.handler = handler

        Timber.i("execute is invoked. ${applet.metaData.getTitle()}")

        var outputString = inputJson?.get("inputText") ?: run {
            terminate(ERError("Input is null"))
            return
        }

        preferenceJson?.get("preText")?.let {
            outputString = it.toString() + outputString.toString()
        }
        preferenceJson?.get("postText")?.let {
            outputString = outputString.toString() + it.toString()
        }
        this.output = mapOf("outputText" to outputString.toString())

        // Get the current queue and make the request based on the queue
        val queue = QCC5100Queue.current ?: run {
            terminate(ERError("No queue to send the request"))
            return
        }
        val request: BleOperation = when (queue) {
            CommandReceiverCommand.BUTTON_1_RESULT -> {
                BleApi.SendResultButton1(message = outputString.toString())
            }
            CommandReceiverCommand.BUTTON_2_RESULT -> {
                BleApi.SendResultButton2(message = outputString.toString())
            }
            CommandReceiverCommand.DISPLAY -> {
                BleApi.SendDisplayData(message = outputString.toString())
            }
            else -> {
                terminate(ERError("Invalid queue found"))
                return
            }
        }

        // Stop running if the response is not received within a pre-defined timeout period.
        backgroundHandler.postDelayed({
            if (!isRunning) {
                return@postDelayed
            } else {
                terminate(ERError("No response within timeout period($timeoutPeriod ms)"))
            }
        }, timeoutPeriod)

        // Toggle the flag to indicate that the request is running & Set the callback
        isRunning = true
        BleGattCallback.sharedInstance?.setCallback(this)

        // Send the request
        BleCentralClient.sharedInstance?.request(request) ?: run {
            terminate(ERError("BleCentralClient.sharedInstance is null"))
        }
    }

    private fun terminate(error: ERError? = null) {
        error
            ?.let { Timber.e("terminate() [ERR] ${it.errorMessage}") }
            ?: run { Timber.d("terminate() [INF] Finished successfully.") }
        isRunning = false
        BleGattCallback.sharedInstance?.removeCallback(this)
        externalScope.launch {
            handler?.completed(output, error)
        }
    }

    ///
    /// Overrides of IBleCentralCallback
    ///
    override fun onCharacteristicChanged(
        gatt: BluetoothGatt,
        characteristic: BluetoothGattCharacteristic,
        value: ByteArray
    ) {
        val receivedPayload = value.toSlateRiiiverPayload()
        when (val w2pCommand = receivedPayload.commandId.toInt().onBoardUserEventCommand()) {
            OnBoardUserEventCommand.SUCCESS -> {
                Timber.d("onCharacteristicChanged() [INF] commandId:${w2pCommand.rawValue} = ${w2pCommand.name}")
                terminate()
            }
            OnBoardUserEventCommand.FAILURE -> {
                Timber.d("onCharacteristicChanged() [INF] commandId:${w2pCommand.rawValue} = ${w2pCommand.name}")
                terminate(ERError("Received FAILURE from the slate"))
            }
            else -> {
                Timber.d("onCharacteristicChanged() [INF] Not handled commandId: ${receivedPayload.commandId}")
            }
        }
    }
}